"""
====================
 PyS60RemoteControl
====================
http://pys60rc.sourceforge.net/
(c) 2007 Dorian Scholz
Released under the GNU General Public Licence
"""
import glob, wx, sys, os, os.path, socket
sys.path.append('lib')
sys.path.append('phone')
sys.path.append('phone' + os.sep + 'libs')
sys.path.append('plugins')
import Skin, BluetoothSerial, Configuration, Logger

class PyS60RemoteControl():
  """
    Main class
  """
  def __init__(self):
    # simple log function
    self.logger = Logger.Logger()
    self.log = self.logger.log

    # open main config  
    self.conf = Configuration.Configuration(self.__class__.__name__ + '.conf', self.log)
    
    # read from main config
    self.temp_file_name = self.conf.get('temp_path', 'temp') + os.sep + 'temp.jpg'
    self.image_path = self.conf.get('image_path', 'images')
    self.logo_filepath = self.image_path + os.sep + self.conf.get('logo', 'pys60rc-logo.jpg')
    self.plugin_path = self.conf.get('plugin_path', 'plugins')
    com_port = self.conf.get('com_port', 8)
    verbose = self.conf.get('verbose', 0)
    
    self.logger.set_verbose(verbose)

    # open bluetooth serial port
    self.bluetooth = BluetoothSerial.BluetoothSerial(com_port, self.log)

    self.cache = {}
    self.drawing_queue = []
    self.display_width, self.display_height = 0, 0
    
    # load event dispatcher
    self.event_dispatcher = None
    event_dispatcher_classname = self.conf.get('event_dispatcher', None)
    if event_dispatcher_classname:
      self.event_dispatcher = self._import_file('%s_EventDispatcherPlugin.py' % event_dispatcher_classname)

    # startup graphics support
    wx.PySimpleApp()

    # "skin" class
    self.skin = Skin.Skin(self)

    # load all plugins from files named {PluginClassName}_RemoteControlPlugin.py
    self.plugins = {}
    self.plugin_path = self.conf.get('plugin_path', 'plugins')
    self.load_plugins()

    self.current_plugin = None
    self.menu = self.conf.get('menu', {'dont_add_plugins':[],'structure':[]})
    self.menuentries = self.menu['structure']
    for plugin_class in self.plugins.keys(): # add plugins which are not in the dont_add list
      if not plugin_class in self.menu['dont_add_plugins']:
        self.menuentries.insert(0, (plugin_class, 'plugin ' + plugin_class))
    
  def run(self):
    """ run the mainloop until self.bluetooth.finished is True """
    try:
      while not self.bluetooth.finished:
        line = self.bluetooth.readline().rstrip()
        if line:
          self.log(5, 'line', line)
          self.evaluate(line)
    except socket.error:
      self.log(0, 'socket error:')
    except:
      if self.log(0, 'exception:'):
          import traceback
          traceback.print_exc()
    self.bluetooth.close()
      
  def evaluate(self, line):
    """ evaluate the line received from the phone """
    self.log(4, 'eval', line)
    parts = line.split(' ', 1)
    head = parts[0]
    if not self.display_width and head != 'init': # phone app forgot to send 'init'?
      self.bluetooth.clearBuffer()
      self.bluetooth.send_line('init')
    elif head == 'init': # init display_width display_height\n #phone app coming up
      if self.current_plugin:
        self.current_plugin.stop()
        self.current_plugin = None
      self.display_width = int(parts[1].split(' ')[0])
      self.display_height = int(parts[1].split(' ')[1])
      default_skins = self.conf.get('default_skins', default = [(240, 'n73')])
      default_skins.sort()
      default_skin_name = None
      for skin_width, skin_name in default_skins:
        if skin_width <= self.display_width or not default_skin:
          default_skin_name = skin_name
      self.skin.set_skin(self.conf.get('use_skin', default = default_skin_name))
      self.bluetooth.send_atomic([
        ('line', 'set_menu'),
        ('object', self.menuentries),
      ])
      self.show_logo()
    elif head == 'quit': # quit\n #phone told us to quit
      if self.current_plugin:
        self.current_plugin.stop()
      self.bluetooth.quit()
    elif head == 'set_skin': # set_skin skin_name\n
      skin_name = parts[1]
      self.conf.set('use_skin', skin_name)
      if self.current_plugin:
        self.current_plugin.stop()
      self.skin.set_skin(skin_name)
      if self.current_plugin:
        self.current_plugin.start()
      else:
        self.show_logo()
    elif head == 'stop_plugin': # stop_plugin\n
      if self.current_plugin:
        self.current_plugin.stop()
        self.current_plugin = None
      self.show_logo()
    elif head == 'killboth': # killboth\n #phone app want us to quit and to send a 'quit' command back
      self.log(2, head)
      if self.current_plugin:
        self.current_plugin.stop()
      self.bluetooth.send_line('quit')
      self.bluetooth.quit()
    elif head == 'bye': # bye\n #phone app is going down...
      self.log(2, head)
      if self.current_plugin:
        self.current_plugin.stop()
        self.current_plugin = None
      try:
        self.bluetooth.send_line('bye')
      except socket.error:
        self.log(4, 'could not send "bye"...')
    elif head == 'send_files': # send_files {hash_list}\n #phone app requesting files by hash values
      hash_list = eval(parts[1])
      self.log(2, 'hash_list', hash_list)
      for filehash in hash_list:
        filepath = self.cache.get(filehash)
        if filepath:
          self.send_file(filepath, filehash)
        else:
          self.log(2, 'unknown filehash', filehash, self.cache)
    elif head == 'assign_keys': # assign_keys\n #phone app wants to reassign scancodes to virtual keys
      if self.current_plugin:
        self.current_plugin.assign_keys()
      else:
        self.log(0, 'no plugin loaded to assign keys for.')
    elif head == 'remote_key': #remote_key {key_event}\n # phone app sends us a key_event
      #key_event = self.bluetooth.recv_pyobj()
      key_event = eval(parts[1])
      if self.current_plugin:
        # if the callback returns True, dispatch this key_event
        if self.current_plugin.key_event(key_event):
          event_str = '%(type)01x.%(keycode)04x.%(scancode)02x.%(modifiers)03x' % key_event
          self.dispatch_event(event_str)
      else:
        self.log(4, 'no plugin loaded to handle remote key.')
    elif head == 'plugin': # plugin {plugin_name}\n # phone wants us to start a plugin
      key = parts[1]
      if key in self.plugins:
        if self.current_plugin:
          self.current_plugin.stop()
        self.current_plugin = self.plugins[key]
        self.current_plugin.start()
      else:
        self.log(0, 'unknown plugin: ' + key)
    elif head == 'output': # output {write_buffer} # phone console output
      #output = self.bluetooth.recv_pyobj()
      output = eval(parts[1])
      self.logger.print_to_log(1, 'PHONE CONSOLE OUTPUT', output)
    else: # unknown command
      self.log(2, 'unknown command', parts)
   
  def dispatch_event(self, event_str):
    """ if there is an event_dispatcher, dispatch the event_str """
    if self.event_dispatcher:
      self.event_dispatcher.trigger_event(event_str)

  def show_logo(self):
    self.draw_image_cached(self.skin.image_filepath, bg = True)
    self.restore_bg()
    self.draw_image_cached(self.skin.logo_filepath, bg = False)
    self.draw_on_phone()

  def get_filehash(self, filepath):
    filestat = os.stat(filepath)
    return hash((filepath, filestat.st_size, filestat.st_mtime))

  def send_file(self, filepath, filehash = None):
    if not filehash:
      filehash = self.get_filehash(filepath)
    filename = filepath.split(os.sep)[-1]
    self.bluetooth.send_atomic([
      ('line', 'file %s %d' % (filename, filehash)),
      ('file', filepath),
    ])
    return filehash

  def save_image(self, image, filename, quality = 75):
    """ saves the bitmap to 'filename' 
        @image: wxImage or wxBitmap
        @filename: where to save the image
    """
    if image.ClassName == 'wxBitmap':
      image = image.ConvertToImage()
    ext = filename.split('.')[-1]
    if ext == 'png':
      filetype = wx.BITMAP_TYPE_PNG
    elif ext[:2] == 'jp':
      filetype = wx.BITMAP_TYPE_JPEG
      image.SetOptionInt(wx.IMAGE_OPTION_QUALITY, quality)
    else:
      raise Exception, 'Can only save to PNG or JPEG format! File extention "%s" unsupported!' % ext
    image.SaveFile(filename, filetype)
    return ext

  def show_image(self, image, x = 0, y = 0, quality = 75):
    """ saves the image to a temp file, sends it to the phone and shows it """
    filepath = self.temp_file_name
    filename = filepath.split(os.sep)[-1]
    self.save_image(image, filepath, quality)
    self.bluetooth.send_atomic([
      ('line', 'draw_image %s %s %s' % (filename, x, y)), 
      ('file', filepath),
    ])
  
  def draw_image_cached(self, *args, **kwargs):
    filepath = args[0]
    filehash = self.get_filehash(filepath)
    self.cache[filehash] = filepath
    args = (filehash,) + args[1:]
    self.draw_image(*args, **kwargs)
  
  def set_font(self, *args, **kwargs):
    self.drawing_queue.append(('set_font', args, kwargs))

  def draw_text(self, *args, **kwargs):
    self.drawing_queue.append(('draw_text', args, kwargs))

  def draw_image(self, *args, **kwargs):
    self.drawing_queue.append(('draw_image', args, kwargs))

  def restore_bg(self, *args, **kwargs):
    self.drawing_queue.append(('restore_bg', args, kwargs))

  def draw_on_phone(self):
    if self.drawing_queue:
      self.bluetooth.send_atomic([
        ('line', 'draw'), 
        ('object', self.drawing_queue),
      ])
      self.drawing_queue = []

  def _import_file(self, filename):
    """ 
      @filename : 'ClassName_Type.py'
      imports the module 'filename' and returns an instance of class 'ClassName'
    """
    module_name = filename.rsplit('.')[0] # cut of the extension
    class_name = module_name.rsplit('_')[0] # the class_name is everything left of the last '_'
    try:
      module = __import__(module_name, globals(), locals(), [], -1)
      class_instance = eval('module.%s(self)' % (class_name))
    except ImportError, ex:
      self.log(0, 'could not import class %s from module %s' % (class_name, module_name))
      self.log(0, str(ex))
      return None
    return class_instance

  def load_plugins(self):
    """ load RemoteControlPlugins from files named {PluginClassName}_RemoteControlPlugin.py """
    self.plugins = {}
    sys.path.append(self.plugin_path)
    for plugin_filename in glob.glob(self.plugin_path + os.sep + '*_RemoteControlPlugin.py'):
      self.log(4, 'plugin file found: %s' % plugin_filename)
      plugin_filename = os.path.split(plugin_filename)[1]
      class_instance = self._import_file(plugin_filename)
      if class_instance:
        self.log(4, 'imported plugin: %s' % class_instance.get_plugin_name())
        self.plugins[class_instance.__class__.__name__] = class_instance
    

if __name__ == '__main__':
  # init main class
  pyS60RemoteControl = PyS60RemoteControl()
    
  # start main loop
  pyS60RemoteControl.run()
