"""
====================
 PyS60RemoteControl
====================
http://pys60rc.sourceforge.net/
(c) 2007 Dorian Scholz
Released under the GNU General Public Licence
"""
import os
from Plugin import Plugin
from TimerRepeat import TimerRepeat

class RemoteControlPlugin(Plugin):
  """
    Base class for all remote control plugins.
    It offers the following functionality:
      - key learning by pressing a key on the phone
      - calling of key specific commands, when key_events arrive from the phone
      - displaying of the key labels on the phone
      - timed callback of a function for update of the display
  """
  
  key_event_types = {
    'repeat' : 1,
    'up' : 2,
    'down' : 3,
  }
  
  def __init__(self, rc, timer_callback = None, interval = 0.1):
    Plugin.__init__(self, rc)
    self.timer_callback = timer_callback
    self.interval = interval

    self.running = False
    self.timer = None
    self.key_event = self.key_dispatcher # start with the key_dispatcher as key_event handler

    # load values from global config
    self.scancode2key = self.rc.conf.get('scancode2key', default = {})

    # load values from plugin config
    self.commands = self.conf.get('commands', default = {})
    self.key2command = {} # dictonary linking a virtual key to a command_name
    for command_name, command in self.commands.items():
      if command.has_key('key'):
        self.key2command[command['key']] = command_name

  def __del__(self):
    self.stop()

  def start(self):
    self.rc.log(2, 'starting plugin ' + self.get_plugin_name())
    self.rc.draw_image_cached(self.rc.skin.image_filepath, bg = True)
    self.rc.restore_bg()
    self.running = True
    self.show_title(update = False)
    self.show_keymap()
    if self.timer_callback:
      self.timer = TimerRepeat(self.interval, self.timer_callback)
      self.timer.start()

  def stop(self):
    self.rc.log(2, 'stopping plugin ' + self.get_plugin_name())
    self.running = False
    if self.timer:
      self.timer.cancel() # stop repeating
      self.timer.join()   # wait for thread to finish

  def show_title(self, update = True):
    x = int((self.rc.skin.title_rect[2] - self.rc.skin.title_rect[0]) / 2)
    y = int((self.rc.skin.title_rect[3] - self.rc.skin.title_rect[1]) / 2)
    text_position_list = [(self.get_plugin_name(), x, y, 'center', 'middle')]
    self.rc.set_font(**self.rc.skin.title_font)
    self.rc.restore_bg(*self.rc.skin.title_rect)
    self.rc.draw_text(text_position_list)
    if update:
      self.rc.draw_on_phone()

  def show_keymap(self, update = True):
    text_position_list = []
    text_position_list2 = []
    for key in self.rc.skin.key_coords.keys():
      if self.key2command.has_key(key):
        x, y = self.rc.skin.key_coords[key]
        try:
          text = self.commands[self.key2command[key]]['short']
        except:
          text = self.key2command[key]
        text_position_list.append((text, x, y, 'center', 'middle'))
        text_position_list2.append((text, x + 1, y + 1, 'center', 'middle'))
    self.rc.restore_bg(*self.rc.skin.keypad_rect)
    self.rc.set_font(**self.rc.skin.keypad_font)
    self.rc.draw_text(text_position_list)
    self.rc.set_font(**self.rc.skin.keypad_font2)
    self.rc.draw_text(text_position_list2)
    if update:
      self.rc.draw_on_phone()
    
  def show_text(self, text, x = None, y = None, update = True):
    if x == None:
      x = self.rc.skin.display_rect[0]
    if y == None:
      y = int((self.rc.skin.display_rect[3] - self.rc.skin.display_rect[1]) / 2)
    self.rc.set_font(**self.rc.skin.display_font)
    self.rc.restore_bg(*self.rc.skin.display_rect)
    text_position_list = [(text, x, y, 'left', 'middle')]
    self.rc.draw_text(text_position_list)
    if update:
      self.rc.draw_on_phone()

  def assign_keys(self):
    if len(self.scancode2key) > 0:
      if len(self.commands) > 0:
        self.stop()
        self.key2command = {}
        self.command_list = self.commands.keys()
        self.command_list.sort()
        self.key_event = self.key_learner

        self.key_learner(None)
      else:
        self.rc.log(0, 'No commands! Please check the plugin conf!')
    else:
      self.rc.log(0, 'No keymap! Please use the KeymapLearner plugin first...')

  def key_learner(self, key_event):
    if key_event:
      if key_event['type'] != RemoteControlPlugin.key_event_types['down']: # only key_down events are used to identify the key
        return False
      # assign this key to the current command
      scancode = key_event['scancode']
      if self.scancode2key.has_key(scancode):
        virtual_key = self.scancode2key[scancode]
        command = self.command_list.pop(0)
        self.key2command[virtual_key] = command
        self.commands[command]['key'] = virtual_key

    # show next command to assign a key to
    if self.command_list:
      command = self.command_list[0]
      self.rc.log(3, command, self.commands[command])
      if self.commands[command].has_key('description'):
        description = self.commands[command]['description']
      else:
        description = command
      self.rc.log(2, 'Assigning keys: Press key for: ' + description)
      self.show_keymap()
      self.show_text('Assigning keys\nPress key for:\n' + description)
    else: # done assigning keys
      self.conf.set('commands', self.commands)
      self.show_keymap()
      self.show_text('Finished assigning keys.')
      self.key_handler = self.key_dispatcher
      self.start()

    return False # while learning never dispatch the key_event

  def key_dispatcher(self, key_event):
    if key_event['type'] == RemoteControlPlugin.key_event_types['up']:
      return True # discard key_up events for now, but let them be dispatched
    scancode = key_event['scancode']
    command = None
    if self.scancode2key.has_key(scancode):
      virtual_key = self.scancode2key[scancode]
      if self.key2command.has_key(virtual_key) and self.commands.has_key(self.key2command[virtual_key]):
        command = self.commands[self.key2command[virtual_key]]
    # if no command is defined for this key or the command_callback returns True, dispatch this key_event
    return not command or self.command_dispatcher(command, key_event)

  def command_dispatcher(self, command, key_event):
    if command.has_key('method_name'):
      return eval('self.%s(command, key_event)' % command['method_name'])
    else:
      return self.command_callback(command, key_event)
  
  def command_callback(self, command, key_event):
    """This method should be overridden in the actual plugin class to make use of the command and key_event.
    It will be called for all commands that have not specified a handler method via the 'method_name' entry in the config.
    This dummy handler just returns True, so the key_event is dispatched using the current event_dispatcher."""
    return True

