"""
====================
 PyS60RemoteControl
====================
http://pys60rc.sourceforge.net/
(c) 2007 Dorian Scholz
Released under the GNU General Public Licence
"""
from RemoteControlPlugin import RemoteControlPlugin
import wx, os
from win32api import mouse_event

mouse_event_constants = {
  'MOVE'        : 0x0001, # mouse move 
  'LEFTDOWN'    : 0x0002, # Left button down 
  'LEFTUP'      : 0x0004, # Left button up 
  'RIGHTDOWN'   : 0x0008, # Right button down 
  'RIGHTUP'     : 0x0010, # Right button up 
  'MIDDLEDOWN'  : 0x0020, # middle button down 
  'MIDDLEUP'    : 0x0040, # middle button up 
  'XDOWN'       : 0x0080, # x button down 
  'XUP'         : 0x0100, # x button down 
  'WHEEL'       : 0x0800, # wheel button rolled 
  'VIRTUALDESK' : 0x4000, # map To entire virtual Desktop 
  'ABSOLUTE'    : 0x8000, # absolute move 
}

class Mouse(RemoteControlPlugin):

  def __init__(self, rc):
    RemoteControlPlugin.__init__(self, rc, self.mouse_timer_callback, 0.5)
    self.image_quality = self.conf.get('image_quality', default = 75)
    self.move_speed = self.conf.get('move_speed', default = 5)
    self.max_move_factor = self.conf.get('max_move_factor', default = 15)
    self.max_wheel_factor = self.conf.get('max_wheel_factor', default = 5)
    self.zoom_level = self.conf.get('zoom_level', default = 1)
    self.rotation = self.conf.get('rotation', default = 0)
    self.mousecursor = wx.Bitmap(self.rc.plugin_path + os.sep + self.conf.get('mouse_cursor', default = 'MouseCursor.png'))
    self.repeat_count = 0

  def start(self):
    self.dispaly_size = (self.rc.display_width, self.rc.display_height)
    self.change_view()
    RemoteControlPlugin.start(self)

  def stop(self):
    RemoteControlPlugin.stop(self)
    self.conf.update({
      'image_quality': self.image_quality,  # remember image_quality
      'move_speed': self.move_speed,        # remember move_speed
      'zoom_level': self.zoom_level,        # remember zoom_level
      'rotation': self.rotation,            # remember rotation
    })

  def change_view(self, command = None, key_event = None):
    if command:
      if command.has_key('zoom') and self.zoom_level + command['zoom'] > 0:
        self.zoom_level += command['zoom']
      if command.has_key('rotate'):
        self.rotation += command['rotate']
        self.rotation %= 4
    self.display_width = self.dispaly_size[self.rotation % 2]
    self.display_height = self.dispaly_size[(self.rotation + 1) % 2]
    self.width = self.display_width * self.zoom_level
    self.height = self.display_height * self.zoom_level
    self.rc.log(3, self.zoom_level, self.width, self.height)
    self.w2 = int(self.width / 2)
    self.h2 = int(self.height / 2)

  def change_image_quality(self, command = None, key_event = None):
    self.image_quality = min(100, max(5, self.image_quality + command['delta']))

  def change_mouse_speed(self, command = None, key_event = None):
    self.move_speed = max(1, self.move_speed + command['delta'])

  def mouse_move(self, command, key_event):
    dx = (int(self.rotation < 2) * 2 - 1) * command['direction'][self.rotation % 2]
    dy = (int(self.rotation % 3 == 0) * 2 - 1) * command['direction'][(self.rotation + 1) % 2]
    if key_event['type'] == RemoteControlPlugin.key_event_types['repeat']:
      self.repeat_count = min(self.repeat_count + 0.2, self.max_move_factor)
    else:
      self.repeat_count = 1
    speed = int(self.repeat_count * max(1, self.move_speed * self.zoom_level))
    self.rc.log(4, dx * speed, dy * speed)
    mouse_event(mouse_event_constants['MOVE'], dx * speed, dy * speed, 0, 0)

  def mouse_wheel(self, command, key_event):
    movement = command['movement']
    self.rc.log(4, movement)
    if key_event['type'] == RemoteControlPlugin.key_event_types['repeat']:
      self.repeat_count = min(self.repeat_count + 0.2, self.max_wheel_factor)
    else:
      self.repeat_count = 1
    movement += int(movement * self.repeat_count)
    mouse_event(mouse_event_constants['WHEEL'], 0, 0, movement, 0)

  def mouse_click(self, command, key_event):
    mouse_event(mouse_event_constants[command['button'].upper() + 'DOWN'], 0, 0, 0, 0)
    for i in range(1000): # wait a few millisecs
      pass
    mouse_event(mouse_event_constants[command['button'].upper() + 'UP'], 0, 0, 0, 0)
        
  def mouse_timer_callback(self):
    point = wx.GetMousePosition()
    capture = self.capture(point.x, point.y) 
    if self.zoom_level != 1 or self.rotation != 0:
      image = capture.ConvertToImage()
      if self.zoom_level != 1:
        if self.zoom_level >= 3: # for big zoom factors use the faster box sampling
#          image = image.Scale(self.display_width, self.display_height)
          image = image.ResampleBox(self.display_width, self.display_height)
        else: # for smaller zoom factors use the high quality bicubic sampling
          image = image.ResampleBicubic(self.display_width, self.display_height)
      if self.rotation != 0:
        if self.rotation % 2 == 1:
          image = image.Rotate90(self.rotation == 1)
        else:
          image = image.Rotate90().Rotate90()
      self.rc.show_image(image, quality = self.image_quality)
    else:
      self.rc.show_image(capture, quality = self.image_quality)
    self.rc.draw_on_phone()

  def capture(self, x, y):
    self.rc.log(4, x, y)
    bitmap = wx.EmptyBitmap(self.width, self.height, -1) # clear 
    memory = wx.MemoryDC()
    memory.SelectObject(bitmap)
    memory.Blit(0, 0, self.width, self.height, wx.ScreenDC(), x - self.w2, y - self.h2)
    memory.DrawBitmap(self.mousecursor, self.w2, self.h2, True)
    memory.SelectObject(wx.NullBitmap)
    return bitmap
